import struct


class Cursor:
    def __init__(self, buffer, offset=0):
        self.buffer = buffer
        if isinstance(buffer, list):
            self.buffer = bytes(buffer)
        self.offset = offset

    def read_bytes(self, size):
        end, total = self.offset + size, len(self.buffer)
        if end > total:
            raise IOError("{} exceed buffer size: {}".format(end, total))
        ret = self.buffer[self.offset:end]
        self.offset = end
        return ret

    def peek_bytes(self, size):
        end, total = self.offset + size, len(self.buffer)
        if end > total:
            raise IOError("{} exceed buffer size: {}".format(end, total))
        return self.buffer[self.offset:end]

    def is_finished(self):
        return self.offset == len(self.buffer)



from canoser import *
import pdb
import pytest

def test_str_to_int():
    with pytest.raises(Exception):
        Uint8.int_unsafe("")
    assert 0 == Uint8.int_unsafe("0")
    assert 0 == Uint8.int_unsafe(b"0")
    assert 1 == Uint8.int_unsafe(b"01")
    assert 0 == Uint8.int_unsafe("0"*100)
    assert 255 == Uint8.int_unsafe("255")
    assert 255 == Uint8.int_unsafe("0255")
    with pytest.raises(Exception):
        Uint8.int_unsafe("-1")
    with pytest.raises(Exception):
        Uint8.int_unsafe("256")

def test_str_to_int_strict():
    assert 0 == Uint8.int_safe("0")
    assert 255 == Uint8.int_safe("255")
    assert -128 == Int8.int_safe("-128")
    assert 127 == Int8.int_safe("127")
    assert 65535 == Uint16.int_safe("65535")
    with pytest.raises(Exception):
        Uint8.int_safe("")
    with pytest.raises(Exception):
        Uint8.int_safe(b"0")
    with pytest.raises(Exception):
        Uint8.int_safe(b"01")
    with pytest.raises(Exception):
        Uint8.int_safe("0"*100)
    with pytest.raises(Exception):
        Uint8.int_safe("0255")
    with pytest.raises(Exception):
        Uint8.int_safe("-1")
    with pytest.raises(Exception):
        Uint8.int_safe("256")
    with pytest.raises(Exception):
        Int8.int_safe("-129")
    with pytest.raises(Exception):
        Int8.int_safe("128")

def test_int():
    assert Int8.encode(16) == Uint8.encode(16)


def test_uint8():
    assert Uint8.encode(16) == b"\x10"
    assert Uint8.decode_bytes(b"\x10") == 16
    assert Uint8.max_value == 255
    assert Uint8.min_value == 0

def test_uint8_illegal():
    with pytest.raises(Exception):
        Uint8.encode(-1)
    with pytest.raises(Exception):
        Uint8.encode(0.1)
    with pytest.raises(Exception):
        Uint8.encode([0])
    with pytest.raises(Exception):
        Uint8.encode(b'0')

def test_int8():
    assert Int8.encode(16) == b"\x10"
    assert Int8.decode_bytes(b"\x10") == 16
    assert Int8.max_value == 127
    assert Int8.min_value == -128
    assert Int8.encode(-1) ==  b"\xFF"
    assert Int8.decode_bytes(b"\xFF") == -1
    assert Int8.decode_bytes(b"\x80") == -128


def test_uint16():
    assert Uint16.encode(16) == b"\x10\x00"
    assert Uint16.encode(257) == b"\x01\x01"
    assert Uint16.decode_bytes(b"\x01\x01") == 257
    assert Uint16.max_value == 65535
    assert Uint16.min_value == 0

def test_int16():
    assert Int16.encode(16) == b"\x10\x00"
    assert Int16.decode_bytes(b"\x10\x00") == 16
    assert Int16.max_value == 32767
    assert Int16.min_value == -32768
    assert Int16.encode(-1) ==  b"\xFF\xFF"
    assert Int16.decode_bytes(b"\xFF\xFF") == -1
    assert Int16.decode_bytes(b"\x00\x80") == -32768

def test_uint32():
    assert Uint32.encode(16) == b"\x10\x00\x00\x00"
    assert Uint32.encode(0x12345678) == b"\x78\x56\x34\x12"
    assert Uint32.decode_bytes(b"\x78\x56\x34\x12") == 0x12345678


def test_uint64():
    assert Uint64.encode(16) == b"\x10\x00\x00\x00\x00\x00\x00\x00"
    assert Uint64.encode(0x1234567811223344) == b"\x44\x33\x22\x11\x78\x56\x34\x12"
    assert Uint64.decode_bytes(b"\x44\x33\x22\x11\x78\x56\x34\x12" ) == 0x1234567811223344

def test_bool():
    assert BoolT.encode(True) == b"\1"
    assert BoolT.encode(False) == b"\0"
    assert BoolT.decode_bytes(b"\1") == True
    assert BoolT.decode_bytes(b"\0") == False
    with pytest.raises(TypeError):
    	BoolT.decode_bytes("\x02")

def test_bytes():
    assert BytesT.pack(1,2,3) == BytesT.pack(*[1,2,3])
    assert BytesT.pack(1,2,3) == bytes([1,2,3])

def test_array():
    arrt = ArrayT(Uint8, 2)
    assert arrt.encode([1, 2]) == b'\x02\x00\x00\x00\x01\x02'
    arr = arrt.decode(Cursor(b'\x02\x00\x00\x00\x01\x02'))
    assert arr == [1, 2]
    with pytest.raises(TypeError):
        arrt.decode(Cursor(b'\x01\x00\x00\x00\x01\x02'))
    with pytest.raises(TypeError):
        arrt.decode(Cursor(b'\x03\x00\x00\x00\x01\x02'))

def test_deserialize_int_array():
    arrt = ArrayT(BoolT, 2)
    bools = arrt.decode(Cursor([2,0,0,0,1,0]))
    assert bools == [True, False]

def test_tuple():
    tuplet = TupleT(StrT, Uint8, BoolT)
    assert tuplet.encode(("abc", 1, False)) == b'\x03\x00\x00\x00\x61\x62\x63\x01\x00'
    ret = tuplet.decode(Cursor(b'\x03\x00\x00\x00\x61\x62\x63\x01\x00'))
    assert ret == ("abc", 1, False)




from canoser.types import (  # noqa: F401
    StrT, BytesT, BoolT, ArrayT, MapT, TupleT,
    Uint8, Uint16, Uint32, Uint64, Int8, Int16, Int32, Int64
)
from canoser.cursor import Cursor  # noqa: F401
from canoser.struct import Struct  # noqa: F401
from canoser.rust_enum import RustEnum  # noqa: F401
from canoser.rust_optional import RustOptional  # noqa: F401
from canoser.delegate_t import DelegateT  # noqa: F401
from canoser.util import bytes_to_int_list, hex_to_int_list  # noqa: F401



import struct
from canoser.base import Base


class IntType:
    _pack_map = {8: "B", 16: "H", 32: "L", 64: "Q"}

    def __init__(self, bits, signed=False):
        self.bits = bits
        self.signed = signed

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        if self.signed:
            return f"Int{self.bits}"
        else:
            return f"Uint{self.bits}"

    @classmethod
    def pretty_print_obj(cls, value, buffer, ident):
        buffer.write(f'{value}')

    def pack_str(self):
        endian = '<'
        pack = IntType._pack_map[self.bits]
        if self.signed:
            pack = pack.lower()
        return endian + pack

    def encode(self, value):
        return struct.pack(self.pack_str(), value)

    def decode_bytes(self, bytes):
        return struct.unpack(self.pack_str(), bytes)[0]

    def decode(self, cursor):
        bytes = cursor.read_bytes(int(self.bits / 8))
        return self.decode_bytes(bytes)

    @property
    def max_value(self):
        if self.signed:
            return 2**(self.bits - 1) - 1
        else:
            return 2**(self.bits) - 1

    @property
    def min_value(self):
        if self.signed:
            return -2**(self.bits - 1)
        else:
            return 0

    def int_unsafe(self, s):
        ret = int(s)
        self.check_value(ret)
        return ret

    def int_safe(self, s):
        if not isinstance(s, str):
            raise TypeError(f"{s} is not instance of <str>.")
        if len(s) < 1:
            raise TypeError(f"'{s}' is empty.")
        len_min = len(str(self.min_value))
        len_max = len(str(self.max_value))
        if len(s) > max(len_min, len_max):
            raise TypeError(f"Length of {s} is larger than max:{max(len_min, len_max)}.")
        ret = int(s)
        self.check_value(ret)
        return ret

    def check_value(self, value):
        if not isinstance(value, int):
            raise TypeError(f"{value} is not instance of <int>.")
        min, max = self.min_value, self.max_value
        if value < min or value > max:
            raise TypeError('value {} not in range {}-{}'.format(value, min, max))


Int8 = IntType(8, True)
Int16 = IntType(16, True)
Int32 = IntType(32, True)
Int64 = IntType(64, True)

Uint8 = IntType(8)
Uint16 = IntType(16)
Uint32 = IntType(32)
Uint64 = IntType(64)


class StrT:
    @classmethod
    def encode(self, value):
        output = b''
        utf8 = value.encode('utf-8')
        output += Uint32.encode(len(utf8))
        output += utf8
        return output

    @classmethod
    def decode(self, cursor):
        strlen = Uint32.decode(cursor)
        return str(cursor.read_bytes(strlen), encoding='utf-8')

    @classmethod
    def check_value(self, value):
        if not isinstance(value, str):
            raise TypeError('value {} is not string'.format(value))

    @classmethod
    def pretty_print_obj(cls, value, buffer, ident):
        buffer.write(f'{value}')


class BytesT:
    @classmethod
    def pack(self, *uint8s):
        output = b''
        for uint8 in uint8s:
            output += struct.pack("<B", uint8)
        return output

    @classmethod
    def encode(self, value):
        output = b''
        output += Uint32.encode(len(value))
        output += value
        return output

    @classmethod
    def decode(self, cursor):
        strlen = Uint32.decode(cursor)
        return cursor.read_bytes(strlen)

    @classmethod
    def check_value(self, value):
        if not isinstance(value, bytes):
            raise TypeError('value {} is not bytes'.format(value))

    @classmethod
    def pretty_print_obj(cls, value, buffer, ident):
        buffer.write(f'{value}')

    @classmethod
    def to_json_serializable(cls, obj):
        return obj.hex()



class BoolT:
    @classmethod
    def encode(self, value):
        if value:
            return b'\1'
        else:
            return b'\0'

    @classmethod
    def decode_bytes(self, value):
        if value == b'\0':
            return False
        elif value == b'\1':
            return True
        else:
            raise TypeError("bool should be 0 or 1.")

    @classmethod
    def decode(self, cursor):
        value = cursor.read_bytes(1)
        return self.decode_bytes(value)

    @classmethod
    def check_value(self, value):
        if not isinstance(value, bool):
            raise TypeError('value {} is not bool'.format(value))

    @classmethod
    def pretty_print_obj(cls, value, buffer, ident):
        buffer.write(f'{value}')


class ArrayT:

    def __init__(self, atype, fixed_len=None, encode_len=True):
        self.atype = atype
        if fixed_len is not None and fixed_len <= 0:
            raise TypeError("arr len must > 0".format(fixed_len))
        if fixed_len is None and not encode_len:
            raise TypeError("variable length sequences must encode len.")
        self.fixed_len = fixed_len
        self.encode_len = encode_len

    def encode(self, arr):
        output = b""
        if self.encode_len:
            output += Uint32.encode(len(arr))
        for item in arr:
            output += self.atype.encode(item)
        return output

    def decode(self, cursor):
        arr = []
        if self.encode_len:
            size = Uint32.decode(cursor)
            if self.fixed_len is not None and size != self.fixed_len:
                 raise TypeError(f"{size} is not equal to predefined value: {self.fixed_len}")
        else:
            size = self.fixed_len
        for _ in range(size):
            arr.append(self.atype.decode(cursor))
        return arr

    def check_value(self, arr):
        if self.fixed_len is not None and len(arr) != self.fixed_len:
            raise TypeError("arr len not match: {}-{}".format(len(arr), self.fixed_len))
        if not isinstance(arr, list):
            raise TypeError(f"{arr} is not a list.")
        for item in arr:
            self.atype.check_value(item)

    def __eq__(self, other):
        if not isinstance(other, ArrayT):
            return False
        return self.atype == other.atype

    def pretty_print_obj(cls, obj, buffer, ident):
        if cls.atype == Uint8:
            hex = struct.pack("<{}B".format(len(obj)), *obj).hex()
            buffer.write(hex)
        else:
            prefix_blank = '  '
            buffer.write('[\n')
            ident_inner = ident+1
            for _, item in enumerate(obj):
                buffer.write(prefix_blank*ident_inner)
                cls.atype.pretty_print_obj(item, buffer, ident_inner)
                buffer.write(',\n')
            buffer.write(prefix_blank*ident)
            buffer.write(']')

    def to_json_serializable(cls, obj):
        if cls.atype == Uint8:
            return struct.pack("<{}B".format(len(obj)), *obj).hex()
        ret = []
        for _, item in enumerate(obj):
            data = Base.to_json_data(value_type=(item, cls.atype))
            ret.append(data)
        return ret

class MapT:

    def __init__(self, ktype, vtype):
        self.ktype = ktype
        self.vtype = vtype

    def encode(self, kvs):
        output = b""
        output += Uint32.encode(len(kvs))
        odict = {}
        for k, v in kvs.items():
            odict[self.ktype.encode(k)] = self.vtype.encode(v)
        for name in sorted(odict.keys()):
            output += name
            output += odict[name]
        return output

    def decode(self, cursor):
        kvs = {}
        size = Uint32.decode(cursor)
        for _ in range(size):
            k = self.ktype.decode(cursor)
            v = self.vtype.decode(cursor)
            kvs[k] = v
        return kvs

    def check_value(self, kvs):
        if not isinstance(kvs, dict):
            raise TypeError(f"{kvs} is not a dict.")
        for k, v in kvs.items():
            self.ktype.check_value(k)
            self.vtype.check_value(v)

    def __eq__(self, other):
        if not isinstance(other, MapT):
            return False
        return self.ktype == other.ktype and self.vtype == other.vtype

    def pretty_print_obj(cls, obj, buffer, ident):
        prefix_blank = '  '
        buffer.write('{\n')
        ident_inner = ident+1
        for k,v in obj.items():
            buffer.write(prefix_blank*ident_inner)
            buffer.write(f'{k}: ')
            cls.vtype.pretty_print_obj(v, buffer, ident_inner)
            buffer.write(',\n')
        buffer.write(prefix_blank*ident)
        buffer.write('}')

    def to_json_serializable(cls, obj):
        amap = {}
        for k,v in obj.items():
            kk = Base.to_json_data(value_type=(k, cls.ktype))
            vv = Base.to_json_data(value_type=(v, cls.vtype))
            amap[kk] = vv
        return amap

class TupleT:

    def __init__(self, *ttypes):
        self.ttypes = ttypes

    def encode(self, value):
        output = b""
        zipped = zip(self.ttypes, value)
        for k, v in zipped:
            output += k.encode(v)
        return output

    def decode(self, cursor):
        arr = []
        for k in self.ttypes:
            arr.append(k.decode(cursor))
        return tuple(arr)

    def check_value(self, value):
        if len(value) != len(self.ttypes):
            raise TypeError(f"{len(value)} is not equal to {len(self.ttypes)}")
        if not isinstance(value, tuple):
            raise TypeError(f"{value} is not a tuple.")
        zipped = zip(self.ttypes, value)
        for k, v in zipped:
            k.check_value(v)

    def __eq__(self, other):
        if not isinstance(other, TupleT):
            return False
        zipped = zip(self.ttypes, other.ttypes)
        for t1, t2 in zipped:
            if t1 != t2:
                return False
        return True

    def pretty_print_obj(cls, obj, buffer, ident):
        prefix_blank = '  '
        buffer.write('(\n')
        ident_inner = ident+1
        zipped = zip(cls.ttypes, obj)
        for k, v in zipped:
            buffer.write(prefix_blank*ident_inner)
            if issubclass(k, Base):
                buffer.write(f'{k.__name__} ')
            k.pretty_print_obj(v, buffer, ident_inner)
            buffer.write(',\n')
        buffer.write(prefix_blank*ident)
        buffer.write(')')

    def to_json_serializable(cls, obj):
        ret = []
        #https://stackoverflow.com/questions/15721363/preserve-python-tuples-with-json
        #If need to deserialize tuple back later, above link will help.
        zipped = zip(cls.ttypes, obj)
        for k, v in zipped:
            data = Base.to_json_data(value_type=(v, k))
            ret.append(data)
        return ret


def my_import(name):
    components = name.split('.')
    mod = __import__(components[0])
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod

def type_mapping(field_type):
    if field_type == str:
        return StrT
    elif field_type == bytes:
        return BytesT
    elif field_type == bool:
        return BoolT
    elif type(field_type) == list:
        if len(field_type) == 0:
            return ArrayT(Uint8)
        elif len(field_type) == 1:
            item = field_type[0]
            return ArrayT(type_mapping(item))
        elif len(field_type) == 2:
            item = field_type[0]
            size = field_type[1]
            return ArrayT(type_mapping(item), size)
        elif len(field_type) == 3:
            item = field_type[0]
            size = field_type[1]
            encode_len = field_type[2]
            return ArrayT(type_mapping(item), size, encode_len)
        else:
            raise TypeError("Array has one item type, no more.")
        raise AssertionError("unreacheable")
    elif type(field_type) == dict:
        if len(field_type) == 0:
            ktype = BytesT
            vtype = [Uint8]
        elif len(field_type) == 1:
            ktype = next(iter(field_type.keys()))
            vtype = next(iter(field_type.values()))
        else:
            raise TypeError("Map type has one item mapping key type to value type.")
        return MapT(type_mapping(ktype), type_mapping(vtype))
    elif type(field_type) == tuple:
        arr = []
        for item in field_type:
            arr.append(type_mapping(item))
        return TupleT(*arr)
    elif type(field_type) == str:
        return my_import(field_type)
    else:
        return field_type


from canoser.cursor import Cursor
from io import StringIO
import json

class Base:

    def serialize(self):
        return self.__class__.encode(self)

    @classmethod
    def deserialize(cls, buffer, check=True):
        cursor = Cursor(buffer)
        ret = cls.decode(cursor)
        if not cursor.is_finished() and check:
            raise IOError("bytes not all consumed:{}, {}".format(
                len(buffer), cursor.offset))
        return ret

    def __str__(self):
        concat = StringIO()
        self.__class__.pretty_print_obj(self, concat, 0)
        return concat.getvalue()

    def to_json(self, sort_keys=False, indent=4):
        amap = self.to_json_serializable()
        return json.dumps(amap, sort_keys=sort_keys, indent=indent)

    @staticmethod
    def to_json_data(base=None, value_type=None):
        if base is not None:
            return base.to_json_serializable()
        value, atype = value_type
        if hasattr(value, "to_json_serializable"):
            return value.to_json_serializable()
        if hasattr(atype, "to_json_serializable"):
            return atype.to_json_serializable(value)
        return value